#!/usr/bin/python
import argparse
import numpy as np
import seaborn as sns
import matplotlib

matplotlib.use('Agg')
import matplotlib.pyplot as plt
import random, multiprocessing, os
from statsmodels.stats.multitest import multipletests

class globalPar:
    stepFactor = 0.1   # step is 0.1*resolution, hic: 0.1
    iterN = 1000    # iteraction time, hic: 1000
    pvalue = 0.1    # FDR corrected pvalue threshold, hic: 0.1

class Enrichment():
    def __init__(self, bedFILE, resultFILE, resolution, start, end, chr, output=None, plot=False):
        self.chr = chr
        self.resolution = resolution
        self.step = globalPar.stepFactor * self.resolution
        self.start = start
        self.end = end
        self.result = []
        if not plot:
            for i in range(len(resultFILE)):
                result = self.read_result(resultFILE[i])
                self.result.append(result)
        else:
            with open(resultFILE[0], "r") as result:
                lines = result.readlines()
                for line in lines:
                    tad = line.split()
                    self.result.append([[int(tad[3]), int(tad[6])]])
            # print(self.result)
        self.profile = []
        for i in range(len(bedFILE)):
            self.profile.append(self.read_profile(bedFILE[i]))
        self.output = os.path.dirname(output)
        self.enrichment_class = np.zeros(shape=(len(self.result), 3))  # each result: %enrich_K27, %enrich_K36, %no enrich
        self.run()

    def run(self):
        pool = multiprocessing.Pool()
        result = pool.map(self.run_per_result, self.result)
        pool.close()
        pool.join()
        self.enrichment_class = np.array(result)
        # print(self.enrichment_class)
        # np.savetxt(self.output + "/enrichment_" + self.chr + "histonemark.txt", self.enrichment_class, fmt='%.4f')

    def read_profile(self, bedFILE):
        binN = int((self.end - self.start) / self.step)
        profile = np.zeros(int(binN))
        with open(bedFILE, "r") as bed:
            lines = bed.readlines()
            for peak in lines:
                peak = peak.split()
                peak_start = int(peak[1])
                peak_end = int(peak[2])
                if peak[0] == self.chr:
                    length = peak_end - peak_start
                    fc = float(peak[3])
                    if peak_start >= self.start and peak_end <= self.end:
                        left = int((peak_start - self.start) / self.step)
                        right = int((peak_end - self.start) / self.step)
                        if left == right:
                            profile[left] += fc * length
                        else:
                            left_part = self.start + self.step * right - peak_start
                            profile[left] += fc * left_part
                            for mid in range(right - left - 1):
                                profile[left + mid + 1] += fc * self.step
                            profile[right] += fc * (length - left_part - (right - left - 1) * self.step)
                    elif peak_start < self.start and peak_end > self.start:
                        right = int((peak_end - self.start) / self.step)
                        if right == 0:
                            profile[0] += fc * (peak_end - self.start)
                        else:
                            for mid in range(right):
                                profile[mid] += fc * self.step
                            profile[right] += fc * (peak_end - right * self.step - self.start)
                    elif peak_start < self.end and peak_end > self.end:
                        left = int((peak_start - self.start) / self.step)
                        if left == binN - 1:
                            profile[left] += fc * (self.end - peak_start)
                        else:
                            for mid in range(binN - 1, left, -1):
                                profile[mid] += fc * self.step
                            profile[left] += fc * (self.end - (binN - 1 - left) * self.step - peak_start)

        return profile / self.step

    def read_result(self, resultFILE):
        """
        Derive the list of TADs.
        :param resultFILE:
        :return:
        """
        tad_list = []
        with open(resultFILE, "r") as result:
            lines = result.readlines()
            for line in lines:
                tad = line.split()
                if int(tad[3]) < int(tad[6]):
                    tad_list.append([int(tad[3]), int(tad[6])])
                elif int(tad[2]) < int(tad[7]):
                    tad_list.append([int(tad[2]), int(tad[7])])
                else:
                    print(tad)
            # tad_list.append([int(tad[1])*self.resolution, int(tad[5])*self.resolution-self.resolution])
        return tad_list

    def run_per_result(self, result):
        tadN = len(result)

        avg_lr = []
        # calculate the observed average LR value
        for tad in result:
            tad_size = tad[1] - tad[0]
            start_bin = int((tad[0] - self.start) / self.step)
            avg_lr.append(self.calculate_aver_lr(self.profile, start_bin, tad_size))

        # Shuffle intervals
        shuff_lr = []
        ori_list = list(range(len(self.profile[0])))
        for time in range(globalPar.iterN):  # Shuffle times
            shuf_pos = random.sample(ori_list, len(ori_list))
            new_profile = []
            new_profile.append(self.profile[0][shuf_pos])
            new_profile.append(self.profile[1][shuf_pos])
            for tad in result:
                tad_size = tad[1] - tad[0]
                start_bin = int((tad[0] - self.start) / self.step)
                shuff_lr.append(self.calculate_aver_lr(new_profile, start_bin, tad_size))
        median_shuf_lr = np.median(shuff_lr)

        # calculate pvalue per tad
        pvalue = np.zeros(tadN)
        tad_state = np.zeros(tadN)
        for tad in range(tadN):
            query = avg_lr[tad]
            if query >= median_shuf_lr:
                pvalue[tad] = np.sum(shuff_lr >= query) / len(shuff_lr)
                tad_state[tad] = 1  # enriched for K27_value
            else:
                pvalue[tad] = np.sum(shuff_lr <= query) / len(shuff_lr)
                tad_state[tad] = -1 # enriched for K36_value
        # print(query, median_shuf_lr, pvalue[tad])

        # correction using BH
        reject, qvalue, _, _ = multipletests(pvalue, method='fdr_bh', alpha=globalPar.pvalue)
        for tad in range(len(reject)):
            if reject[tad] == False:
                tad_state[tad] = 0
        return [len(tad_state[tad_state == 1]) / tadN, len(tad_state[tad_state == -1]) / tadN, len(tad_state[tad_state == 0]) / tadN]

    def Output(self):
        return self.enrichment_class[:, 0] + self.enrichment_class[:, 1]

    def calculate_aver_lr(self, profile, start_bin, tad_size):
        interval_list = []
        for i in range(int(tad_size / self.step)):
            K27_value = profile[0][start_bin + i]
            K36_value = profile[1][start_bin + i]
            if K27_value > 0 and K36_value > 0:
                interval_list.append(np.log10(K27_value / K36_value))
        try:
            avg_lr = np.sum(interval_list) * self.step / tad_size
        except:
            print("error::", start_bin, tad_size, interval_list)
            avg_lr = 0
        return avg_lr


def main():
    print("blocksize={}, iterN={}, pvalue={}".format(globalPar.stepFactor, globalPar.iterN, globalPar.pvalue))
    parser = argparse.ArgumentParser()
    parser.add_argument("--bed", dest="bed", nargs="+", required=True, help="Epigenome bedgraph file, eg. H3K*me3")
    parser.add_argument("-i", dest="result", nargs="+", required=True, help="TADcaller result file in 8col format")
    parser.add_argument("-r", dest="resolution", type=int, required=True)
    parser.add_argument("-c", dest="chrom", required=True)
    parser.add_argument("-o", dest="output", default=None)
    parser.add_argument("--plot", dest="plot", action="store_true", help="if given, draw the bound profile plot")
    parser.add_argument("-s", dest="start", type=int, default=0, required=True, help="the start bp position")
    parser.add_argument("-e", dest="end", type=int, default=np.inf, required=True, help="the end bp position")
    parser.add_argument("--list", dest="list", action="store_true", help="if given, the input is a list of results")
    parser.add_argument("--suffix", dest="suffix", default=".multi2D_AllH2.tsv", help="default: .binary.original.tsv")
    args = parser.parse_args()

    if args.output == None:
        output = os.path.splitext(args.result[0])[0]
    else:
        output = args.output
    if args.list == True:
        resultList = []
        for line in open(args.result[0], "r"):
            line = line.strip("\n")
            resultList.append(line + args.suffix)
    else:
        resultList = []
        for i in args.result:
            resultList.append(i+args.suffix)
    Enrichment(args.bed, resultList, args.resolution, args.start, args.end, args.chrom, output, args.plot)


if __name__ == "__main__":
    main()



import networkx as nx
import matplotlib, os
matplotlib.use('Agg')
import matplotlib.pyplot as plt

class TreeNode():
    def __init__(self, start, end, indicator="F"):
        self.val = [start, end]
        self.info = None    #se
        self.child = []
        self.parent = None
        self.indicator = indicator


class MultiChildTree():
    def __init__(self, start, end, indicator="F"):
        self.root = TreeNode(start, end, indicator)
        self.node_list = []

    def insert(self, start, end, indicator="F"):
        treenode = TreeNode(start, end, indicator)
        self.add(treenode, self.root)

    def add(self, new_node, parent_node):
        current_node = parent_node
        if not parent_node.child:
            parent_node.child.append(new_node)
            new_node.parent = parent_node
        else:
            current_node = parent_node.child[0]
            if (new_node.val[1] < current_node.val[0]): #on the left of current node    #<
                parent_node.child.insert(0, new_node)
                new_node.parent = parent_node
            elif(new_node.val[0] >= current_node.val[0]
                 and new_node.val[1] <= current_node.val[1]):   #child of current node
                self.add(new_node, current_node)
            else:
                for i in range(1, len(parent_node.child), 1):
                    if(new_node.val[0] >= parent_node.child[i].val[0]
                            and new_node.val[1] <= parent_node.child[i].val[1]):
                        self.add(new_node, parent_node.child[i])
                        return
                    elif(new_node.val[0] > parent_node.child[i-1].val[1]
                         and new_node.val[1] < parent_node.child[i].val[0]):    #>, <
                        parent_node.child.insert(i, new_node)
                        new_node.parent = parent_node
                        return
                if(new_node.val[0] > parent_node.child[-1].val[1]): #>
                    parent_node.child.append(new_node)
                    new_node.parent = parent_node
                else:
                    print("node can not insert.", new_node.val[0], new_node.val[1])

        return

    def delete(self, node):
        if not node.child:
            node.parent.child.remove(node)
        else:
            node_pos = node.parent.child.index(node)  #node postion at parent_child list
            for i in node.child:
                i.parent = node.parent
                node.parent.child.insert(node_pos, i)
                node_pos += 1
            node.parent.child.remove(node)
        return

    def pre_traversal(self, tree_root):  # traverse binary tree in pre-order
        if (tree_root != None):
            if tree_root != self.root:
                self.node_list.append(tree_root)
            for i in range(0, len(tree_root.child), 1):
                self.pre_traversal(tree_root.child[i])

    def acquire_list(self):
        self.pre_traversal(self.root)
        return self.node_list

    def find_parent(self, tree_node, bin):
        parent_start = tree_node.val[0]
        parent_end = tree_node.val[1]
        if(len(tree_node.child)!=0 and tree_node.val[0]<=bin and tree_node.val[1]>=bin):
            for i in tree_node.child:
                if(i.val[0]<=bin and i.val[1]>=bin):
                    parent_start, parent_end = self.find_parent(i, bin)
        return parent_start, parent_end

#visualization
def create_graph(G, node, pos={}, x=0, y=0, layer=1):
    key = str(node.val[0])+"-"+str(node.val[1])
    pos[key] = (x, y)
    child_num = len(node.child)
    if(child_num == 1):
        child = node.child[0]
        key_child = str(child.val[0]) + "-" + str(child.val[1])
        G.add_edge(key, key_child)
        (child_x, child_y) = (x, y - 1)
        create_graph(G, child, x=child_x, y=child_y, pos=pos, layer=layer+1)
    else:
        for i in range(0, child_num, 1):
            child = node.child[i]
            key_child = str(child.val[0]) + "-" + str(child.val[1])
            G.add_edge(key, key_child)
            (child_x, child_y) = (x - 1/2**layer + 2*1/2**layer/(child_num-1)*i, y - 1)
            create_graph(G, child, x=child_x, y=child_y, pos=pos, layer=layer+1)
    return (G, pos)

def draw(node, dirnames):
    graph = nx.DiGraph()
    graph, pos = create_graph(graph, node)
    fig, ax = plt.subplots(figsize=(12, 8))
    nx.draw_networkx(graph, pos, ax=ax, node_size=400, fontsize=16)
    plt.savefig(os.path.join(dirnames, "multichild_tree.pdf"), format='pdf')

def scatter_plot(node_list, title=None, ylabel=None, p_parent=None, output=False):
    """
    :param node_list: list
    :param title: string
    :param ylabel: string
    :param p_parent: 'diff' means difference between node and its parent;
                    'quo' means quotient of node and its parent;
                    default means inherent info.
    :return:
    """
    all_boundary = []
    X = range(0, len(node_list))
    Y = []
    Color = []
    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    if(title != None):
        ax1.set_title(title)
    plt.xlabel('Node list with pre-order')
    if(ylabel != None):
        plt.ylabel(ylabel)
    for i in node_list:
        if(p_parent==None):
            Y.append(i.info)
        else:
            if(i.parent==None):
                Y.append(i.info)
            else:
                if(p_parent =='diff'):
                    Y.append(i.info - i.parent.info)
                    if(output==True and (i.info-i.parent.info > node_list[0].info)):
                        all_boundary.append([i.val[0], i.val[1]])   #start and end pos of true node
                elif(p_parent == 'quo'):
                    Y.append(i.info/i.parent.info)
        if(i.indicator == 'T'):
            Color.append('r')
        else:
            Color.append('g')
    ax1.scatter(X, Y, c=Color, marker='o')
    for i in range(0, len(node_list)):
        label = str(node_list[i].val[0]) + '-'+ str(node_list[i].val[1])
        # ax1.annotate(label, (i+1, node_list[i].info))
        ax1.text(i+1, Y[i], label, horizontalalignment='left', size='medium', color=Color[i])
    plt.show()
    if(output==True):
        return all_boundary










#!/usr/bin/python
# -*- coding: utf-8 -*-
import argparse, os, time, multiprocessing
import math
import pulp
import numpy as np
import MultichildTree as MultichildT

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", dest="result", nargs="+", required=True)
    parser.add_argument("-n", dest="count", type=int, required=True)
    parser.add_argument("-b", dest="binInput", action="store_true", help="if given, the input is binList")
    parser.add_argument("--list", dest="listInput", action="store_true", help="if given, the input is a list of results")
    parser.add_argument("--suffix", dest="suffix", default=".binary.original.tsv", help="default: .binary.original.tsv")
    args = parser.parse_args()
    compare_list = []
    if args.listInput == True:
        for line in open(args.result[0], "r"):
            line = line.strip("\n")
            compare_list.append(read_result(line+args.suffix, args.binInput))
    else:
        for i in args.result:
            compare_list.append(read_result(i, args.binInput))
    Compare(compare_list, args.count, suffix=args.suffix)


class Compare():
    def __init__(self, compare_list, binN, suffix="", output_path="."):
        self.n = binN
        self.clusters1 = compare_list[0]    # truth
        self.leaflist1 = self.get_leafList(self.clusters1)
        pool = multiprocessing.Pool()
        result = pool.map(self.run, compare_list)
        pool.close()
        pool.join()
        self.output = np.array(result)
        # np.savetxt(output_path+"/similarity_log_{}.txt".format(suffix), self.output)

    def run(self, cluster):
        self.clusters2 = cluster
        self.leaflist2 = self.get_leafList(self.clusters2)
        overlap = round(self.similarity1(), 4)
        # weighted = round(self.similarity2(), 4)
        # mod = round(self.calculate_moc(), 4)
        nmi = round(self.normMI(), 4)
        # ari = round(self.ARI(), 4)
        # print(overlap, weighted, mod, nmi, ari)
        # return [overlap, weighted, mod, nmi, ari]
        # print(overlap, nmi)
        return [overlap, nmi]

    def Output(self):
        return self.output

    def similarity1(self):
        total_length_1 = len(sum(self.clusters1, []))
        total_length_2 = len(sum(self.clusters2, []))
        # print(total_length_1, total_length_2)
        self.wts = {}
        from_nodes = [k for k in range(1, len(self.clusters1) + 1)]
        to_nodes = [k for k in range(1, len(self.clusters2) + 1)]
        self.ucap = {}
        self.vcap = {}
        for i in range(0, len(from_nodes)):
            self.ucap[from_nodes[i]] = 1
        for i in range(0, len(to_nodes)):
            self.vcap[to_nodes[i]] = 1
        for i in range(0, len(self.clusters1)):
            for j in range(0, len(self.clusters2)):
                interc = list(set(self.clusters1[i]).intersection(set(self.clusters2[j])))
                self.wts[(i + 1, j + 1)] = 2 * len(interc)  # weight assigned when calculating overlapping
        wt = self.create_wt_doubledict(from_nodes, to_nodes)
        p = self.solve_wbm(from_nodes, to_nodes, wt)
        count, selected_edges = self.get_selected_edges(p)
        # print("Sum of wts of selected edges = ", self.print_solution(p) / (total_length_1 + total_length_2))
        # return self.print_solution(p) / count
        return self.print_solution(p) / (total_length_1 + total_length_2)

    # just a convenience function to generate a dict of dicts
    def create_wt_doubledict(self, from_nodes, to_nodes):
        wt = {}
        for u in from_nodes:
            wt[u] = {}
            for v in to_nodes:
                wt[u][v] = 0
        for k, val in self.wts.items():
            u, v = k[0], k[1]
            wt[u][v] = val
        return wt

    def solve_wbm(self, from_nodes, to_nodes, wt):
        ''' A wrapper function that uses pulp to formulate and solve a WBM'''

        prob = pulp.LpProblem("WBM Problem", pulp.LpMaximize)
        # Create The Decision variables
        choices = pulp.LpVariable.dicts("e", (from_nodes, to_nodes), 0, 1, pulp.LpInteger)
        # Add the objective function
        prob += pulp.lpSum([wt[u][v] * choices[u][v]
                            for u in from_nodes
                            for v in to_nodes])  # Calculate the sum of a list of linear expressions
        # Constraint set ensuring that the total from/to each node is less than its capacity
        for u in from_nodes:
            for v in to_nodes:
                prob += pulp.lpSum([choices[u][v] for v in to_nodes]) <= self.ucap[u], ""
                prob += pulp.lpSum([choices[u][v] for u in from_nodes]) <= self.vcap[v], ""
        # The problem data is written to an .lp file
        # prob.writeLP("WBM.lp")
        # The problem is solved using PuLP's choice of Solver
        prob.solve()
        # The status of the solution is printed to the screen
        # print("Status:", pulp.LpStatus[prob.status])
        return prob

    def print_solution(self, prob):
        return round(pulp.value(prob.objective), 4)

    def get_selected_edges(self, prob):
        selected_from = [v.name.split("_")[1] for v in prob.variables() if v.value() > 1e-3]
        selected_to = [v.name.split("_")[2] for v in prob.variables() if v.value() > 1e-3]
        selected_edges = []
        for su, sv in list(zip(selected_from, selected_to)):
            selected_edges.append((su, sv))
            # print(su, sv, self.wts[(int(su), int(sv))])
        return len(selected_edges), selected_edges

    def similarity2(self):
        """
        The weighted similarity between Q and P.
        :param self.clusters1: Q
        :param self.clusters2: P, as the denominator
        :return:
        """
        numerator = 0
        denominator = 0
        for i in range(0, len(self.clusters1)):
            weight_of_everytad1 = []
            for j in range(0, len(self.clusters2)):
                interc = list(set(self.clusters1[i]).intersection(set(self.clusters2[j])))
                simi = len(interc) / (math.sqrt(len(self.clusters1[i]) * len(self.clusters2[j])))
                weight_of_everytad1.append(simi)
            numerator += len(self.clusters1[i]) * max(weight_of_everytad1)
            denominator += len(self.clusters1[i])
        print("weighted similarity = ", numerator / denominator)
        return numerator / denominator

    def get_leafList(self, tad_list):
        data_list_new = sorted(tad_list, key=lambda x: len(x), reverse=True)
        multichildtree = MultichildT.MultiChildTree(1, self.n)
        for i in range(0, len(data_list_new)):
            # print(data_list_new[i][0], data_list_new[i][-1])
            multichildtree.insert(data_list_new[i][0], data_list_new[i][-1], "F")
        node_list = multichildtree.acquire_list()
        leaf_list = []
        total_conv = []
        for i in range(0, len(node_list)):
            if len(node_list[i].child) == 0:
                bin_list = list(range(node_list[i].val[0], node_list[i].val[1] + 1))
                leaf_list.append(bin_list)
                total_conv.extend(bin_list)
        if len(total_conv) != self.n:  # fill in gaps
            # print(len(total_conv), "self.n", self.n)
            covered = True
            gap = []
            for i in range(1, self.n + 1, 1):
                if i not in total_conv and covered == True:
                    gap = []
                    gap.append(i)
                    covered = False
                elif i not in total_conv and covered == False:
                    gap.append(i)
                elif i in total_conv and covered == False:
                    leaf_list.append(gap)
                    total_conv.extend(gap)
                    covered = True
                else:
                    pass
            # print("proof: #bin = {}, #covered bin = {}".format(self.n, len(total_conv)))
            if len(total_conv) > self.n:
                print("&&&&&&&&&&&&&&&&&&&&&&&&&&&&error&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&")
        return leaf_list

    def calculate_moc(self):
        tad_n1 = len(self.leaflist1)
        tad_n2 = len(self.leaflist2)
        if tad_n1 == 1 and tad_n2 == 1:
            return 1
        else:
            sum = 0
            for i in range(tad_n1):
                for j in range(tad_n2):
                    interc = len(set(self.leaflist1[i]).intersection(set(self.leaflist2[j])))
                    sum += pow(interc, 2) / (len(self.leaflist1[i]) * len(self.leaflist2[j]))
            # print("sum = {}".format(sum))
            return (sum - 1) / (np.sqrt(tad_n1 * tad_n2) - 1)

    def normMI(self):
        """
        The normalized mutual information (NMI) is to compare clustering with different numbers of clusters.
        NMI is always a number btw 0 and 1.
        :return: The value of NMI.
        """
        MI = 0
        H1 = 0
        H2 = 0
        time = 0
        for c1 in self.leaflist1:
            len1 = len(list(set(c1)))
            H1 -= (len1/self.n)*np.log2(len1/self.n)
            time += 1
            for c2 in self.leaflist2:
                len2 = len(list(set(c2)))
                if time == 1:
                    H2 -= (len2/self.n)*np.log2(len2/self.n)
                intersection = len(list(set(c1)&set(c2)))
                # print("len1={}, len2={}, inter={}, MI={}".format(len1, len2, intersection, MI))
                if intersection != 0:
                    MI += (intersection/self.n)*np.log2((self.n*intersection)/(len1*len2))
        return MI/np.sqrt(H1 * H2)

    def ARI(self):
        sum_inter = 0
        sum_a = 0
        sum_b = 0
        time = 0
        for c1 in self.leaflist1:
            len1 = len(list(set(c1)))
            sum_a += len1 * (len1 - 1) / 2.0
            time += 1
            for c2 in self.leaflist2:
                len2 = len(list(set(c2)))
                if time == 1:
                    sum_b += len2 * (len2 - 1) / 2.0
                intersection = len(list(set(c1) & set(c2)))
                sum_inter += intersection * (intersection - 1)/ 2.0
        expectedIndex = sum_a*sum_b*2.0/(self.n*(self.n-1))
        return (sum_inter-expectedIndex)/(0.5*(sum_a+sum_b)-expectedIndex)


def read_result(dirnames, binInput=False):
    if os.path.exists(dirnames):
        result_list = []
        with open(dirnames, "r") as result:
            lines = result.readlines()
            for line in lines:
                line = line.split()
                if line != []:
                    if binInput:
                        a = list(map(int, line))
                    else:
                        a = list(range(int(line[1]), int(line[5])+1))
                    result_list.append(a)
        # print("Successfully loading", dirnames, "with", len(result_list), "lines.")
        if len(result_list) == 0:
            return [[0]]
        return result_list
    else:
        return [[0]]


if __name__ == "__main__":
    main()



# _*_ coding: utf-8 _*_
"""
Time:     2021/12/22 11:35
Author:   ZHANG Yuwei
Version:  V 0.1
File:     __init__.py
Describe:
"""
from .main import SupertadSparse
from evaluation import Evaluate

name = 'SuperTLD'


# _*_ coding: utf-8 _*_
"""
Time:     2021/12/13 21:50
Author:   ZHANG Yuwei
Version:  V 0.1
File:     main.py
Describe:
"""

import numpy as np
import pandas as pd
import warnings, subprocess, os
warnings.filterwarnings("ignore")
import smurf
from KR_norm_juicer import KRnorm_asym, KRnorm_sym
from AsymToSym import asym_to_sym
from utils import heatmap


class SupertadSparse():
    def __init__(self, chrom="chr1", resolution=100000, norm=True, smurf=True,
                 supertad="/home/zhangyuwei/hic_TAD/SuperTAD_bakcup/build/SuperTAD", dataweight=0., verbose=False):
        self.chrom = chrom
        self.resolution = resolution
        self.norm = norm
        self.smurf = smurf
        self.supertadPath = supertad
        self.dataweight = dataweight
        self.verbose = verbose

    def pipeline(self, inputMatrix, outpath="./norm_matrix.txt", noise_model="Fano", learningRate=1e-6, lambda_regulation=None, hic=None, alpha=1.):
        if inputMatrix.shape[0] == inputMatrix.shape[1] and np.allclose(inputMatrix, inputMatrix.T):    # check if symmetric
            # symmetric
            if self.verbose:
                heatmap(inputMatrix, "./sym_00_raw")
                heatmap(inputMatrix, "./sym_00_raw_diag", diag=True)

            if self.norm:
                print("Start to perform KR norm on the raw data.")
                imputed_matrix, _ = KRnorm_sym(inputMatrix, oneNorm=False, verbose=self.verbose)
                if self.verbose:
                    np.savetxt("./sym_01_kr.txt", imputed_matrix)
                    heatmap(imputed_matrix, "./sym_01_kr")
                    heatmap(imputed_matrix, "./sym_01_kr_diag", diag=True)
            else:
                imputed_matrix = inputMatrix

            if self.smurf:
                if lambda_regulation:
                    print("Start to impute the symmetric contact map with SMURF with model {} and lambda {}.".format(noise_model, lambda_regulation))
                else:
                    lambda_regulation = 0.01
                    print("Start to impute the symmetric contact map with SMURF with model {} and lambda {}.".format(
                        noise_model, lambda_regulation))
                operator = smurf.SMURF_SYM(n_features=None, lambda2=lambda_regulation, estimate_only=True, noise_model=noise_model,
                                           weight=self.dataweight, verbose=self.verbose, alpha=learningRate)
                imputed_matrix = operator.smurf_impute(pd.DataFrame(imputed_matrix))

                imputed_matrix = imputed_matrix.values
                if self.verbose:
                    np.savetxt("./sym_02_smurf.txt", imputed_matrix)
                    heatmap(imputed_matrix, "./sym_02_smurf")
                    heatmap(imputed_matrix, "./sym_02_smurf_diag", diag=True)
            else:
                imputed_matrix = imputed_matrix

            if hic != None: # data integration
                hicMatrix = np.loadtxt(hic)
                if hicMatrix.shape != imputed_matrix.shape:
                    raise ValueError("The two contact maps' shape doesn't match.")
                outputList = []
                for i in np.arange(0, 1.05, 0.05):
                    alpha = i
                    matrix = alpha * imputed_matrix + (1 - alpha) * hicMatrix
                    np.savetxt(outpath+str(i), matrix)
                    self.Perform_SuperTAD(outpath+str(i))
                    outputList.append(outpath+str(i))
                return outputList, outpath
            else:
                np.savetxt(outpath, imputed_matrix)
                return outpath, self.Perform_SuperTAD(outpath)

        else:   # asymmetric
            if self.smurf:
                if lambda_regulation:
                    print("Start to impute the asymmetric contact map with SMURF with model {} and lambda {}.".format(noise_model, lambda_regulation))
                else:
                    lambda_regulation = 1e-5
                    print("Start to impute the asymmetric contact map with SMURF with model {} and lambda {}.".format(
                        noise_model, lambda_regulation))
                operator = smurf.SMURF(n_features=None, normalize=self.norm, lambda2=lambda_regulation, estimate_only=True,
                                       noise_model=noise_model, weight=self.dataweight, verbose=self.verbose, alpha=learningRate)
                imputed_matrix, G, H = operator.smurf_impute(pd.DataFrame(inputMatrix))
                imputed_matrix = imputed_matrix.values
                if self.verbose:
                    np.savetxt("./asym_smurf.txt", imputed_matrix)
                    heatmap(imputed_matrix, "./asym_smurf")
                    heatmap(imputed_matrix, "./asym_smurf_diag", diag=True)
            else:
                imputed_matrix = inputMatrix

            if hic != None: # data integration
                outputList = []
                for i in np.arange(0, 1.05, 0.05):
                    matrix = asym_to_sym(imputed_matrix, extraNorm=True, hic=hic, alpha=i)
                    np.savetxt(outpath+str(i), matrix)
                    self.Perform_SuperTAD(outpath+str(i))
                    outputList.append(outpath+str(i))
                return outputList, outpath
            else:
                matrix = asym_to_sym(imputed_matrix, extraNorm=True)
                np.savetxt(outpath, matrix)
                result = self.Perform_SuperTAD(outpath)
                return outpath, result

    def Perform_SuperTAD(self, inputPath, hu=1, hd=1, sparse=True):  #sparse mode
        if sparse:
            job = [self.supertadPath, "multi_2d", inputPath, "--hu", str(hu), "--hd", str(hd), "--chrom1", self.chrom, "-r", str(self.resolution), "-s"]
        else:
            job = [self.supertadPath, "multi_2d", inputPath, "--hu", str(hu), "--hd", str(hd), "--chrom1", self.chrom, "-r", str(self.resolution)]
        p = subprocess.Popen(job, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        status = p.wait()
        if sparse:
            return inputPath + ".multi2D_AllH2_sparse.tsv"
        else:
            return inputPath + ".multi2D_AllH2.tsv"




# _*_ coding: utf-8 _*_
"""
Time:     2021/12/14 22:50
Author:   ZHANG Yuwei
Version:  V 0.1
File:     AsymToSym.py
Describe:
"""
import numpy as np


def asym_to_sym(matrix, mode="col", extraNorm=False, hic=None, alpha=1):
    if mode == "row":
        if hic != None:
            hic_matrix = acquire_hic(hic)
            if len(matrix) != len(hic_matrix):
                raise ValueError("RNA-associated data and hic's row length doesn't match.")
            # matrix = np.append(np.sqrt(alpha) * min_max_norm(matrix), np.sqrt(1-alpha)*hic_matrix, axis=1)
            matrix = np.append(np.sqrt(alpha) * matrix, np.sqrt(1 - alpha) * hic_matrix, axis=1)
        oldSum = np.sum(matrix)  # sum of the input matrix
        print("Matrix sum for original matrix: {}".format(oldSum))
        matrix = np.dot(matrix, matrix.T)
    elif mode == "col":
        if hic != None:
            hic_matrix = acquire_hic(hic)
            if len(matrix[0]) != len(hic_matrix):
                raise ValueError("RNA-associated data and hic's column length doesn't match.")
            # matrix = np.append(np.sqrt(alpha) * min_max_norm(matrix), np.sqrt(1-alpha)*hic_matrix.T, axis=0)
            matrix = np.append(np.sqrt(alpha) * matrix, np.sqrt(1 - alpha) * hic_matrix.T, axis=0)
        oldSum = np.sum(matrix)  # sum of the input matrix
        print("Matrix sum for original matrix: {}".format(oldSum))
        matrix = np.dot(matrix.T, matrix)
    else:
        raise ValueError("the setting of mode do not valid")
    if extraNorm:
        norm_matrix = BisectionNorm(matrix, oldSum).norm()
        try:
            if norm_matrix == None:
                print("apply the power to the singular values of symmetric matrix")
                norm_matrix = svd_process(matrix, power=0.5)
        except:
            pass
        matrix = norm_matrix
    else:
        matrix = svd_process(matrix, power=1)
    return matrix


def acquire_hic(hic_path):
    hic_sym = np.loadtxt(hic_path)
    # hic_sym = min_max_norm(hic_sym)
    values, vectors = np.linalg.eigh(hic_sym)   # increasing order
    dim = len(np.where(values > 0)[0])
    values = values[-dim:]
    vectors = vectors[:, -dim:]
    X = np.dot(vectors, np.diag(np.sqrt(values)))  # hic = X.T * X
    print("hic recomb error: {} with shape {}".format(np.sum(hic_sym - np.dot(X, X.T)), X.shape))
    return X


class BisectionNorm():
    def __init__(self, matrix, oldSum):
        self.matrix = matrix
        self.length = len(self.matrix)
        self.oldSum = oldSum
        self.u = None
        self.vt = None
        self.sigma = None

    def norm(self):
        for i in range(self.length):
            self.matrix[i, i] = 0
        # print(np.sum(self.matrix))
        self.u, self.sigma, self.vt = np.linalg.svd(self.matrix)
        alpha = self.cal_val(1e-6, 10, 1e-6)
        print(alpha, "--------------")
        if alpha == None:
            return None
        else:
            new_matrix = np.dot(np.dot(self.u, np.diag(np.power(self.sigma, alpha))), self.vt)
            new_matrix[new_matrix < 0] = 0
            # print(np.sum(new_matrix), oldSum)
            return new_matrix

    def cal_val(self, start, end, precision):
        s = self.func(start)
        e = self.func(end)
        if e * s > 0:
            print("no solution", self.oldSum)
            print("s = {}, e = {}".format(s, e))
            return None
        elif s == 0:
            print("Solution is ", start)
            return start
        elif e == 0:
            print("Solution is ", end)
            return end
        else:
            while abs(end - start) > precision:
                mid = (start + end) / 2.0
                m = self.func(mid)
                if m == 0:
                    print("Solution is ", mid)
                    return mid
                elif s * m < 0:
                    end = mid
                elif m * e < 0:
                    start = mid
            print(start, mid, end)
            return start

    def func(self, alpha):
        matrix = np.dot(np.dot(self.u, np.diag(np.power(self.sigma, alpha))), self.vt)
        for i in range(len(matrix)):
            matrix[i, i] = 0
        matrix[matrix < 0] = 0
        result = np.sum(matrix) - self.oldSum
        # print(alpha, np.sum(matrix), oldSum, result)
        return result


def svd_process(input_matrix, power=1):
    u, sigma, vt = np.linalg.svd(input_matrix)
    length = len(input_matrix)
    tmp_sigma = np.zeros(length)
    for i in range(length):
        if power == -1:
            if (sigma[i] >= 1e-6):
                tmp_sigma[i] = pow(sigma[i], power)
        else:
            try:
                tmp_sigma[i] = pow(sigma[i], power)
            except:
                print("There's no {}th singular value of input matrix.".format(i))
    if power == -1:
        tmp_sigma.sort(reverse=True)

    new_matrix = np.dot(np.dot(u, np.diag(tmp_sigma)), vt)
    new_matrix[np.where(new_matrix < 0)] = 0  # remove negative values from new matrix
    return new_matrix



# _*_ coding: utf-8 _*_
"""
Time:     2021/12/17 17:35
Author:   ZHANG Yuwei
Version:  V 0.1
File:     evaluation.py
Describe:
"""

import numpy as np
from distance_decay import get_pcc, calculate_dis_decay, pearsonr
from SImi_tad import Compare, read_result
from enrichment_boun_foldchange import FoldChange
from enrichment_tad import Enrichment


class Evaluate():
    def __init__(self, chrom="chr1", resolution=100000, hicPath=None, bed=None, bedgraph=None, outputName="evaluation"):
        self.chrom = chrom
        self.resolution = resolution
        self.hicPath = hicPath
        self.Nbin = 0
        self.outputName = outputName
        self.sparse = True
        self.suffix = ".multi2D_AllH2_sparse.tsv"
        self.bed = bed
        self.bedgraph = bedgraph

    def run(self, resultList, outPath="./"):
        self.fileList = resultList
        self.outPath = outPath
        # distance decay
        matrix_simi = self.distance_decay_simi()
        # overlapping ratio, nmi
        simi = self.evaluate_simi()
        # enrichment
        resultList = [self.hicPath+self.suffix]
        for i in self.fileList:
            resultList.append(i + self.suffix)
        bound_enrich = self.evaluate_bound_enrich(resultList)  # each row: fold change, p-value
        tad_enrich = self.evaluate_tad_enrich(resultList)
        resultMatrix = np.concatenate((matrix_simi, simi, bound_enrich, np.reshape(tad_enrich, (-1, 1))), axis=1)
        np.savetxt(self.outPath, resultMatrix)
        return resultMatrix

    def distance_decay_simi(self):
        hic = np.loadtxt(self.hicPath)
        hicDecay = calculate_dis_decay(hic)
        self.Nbin = len(hic)
        matrix_simi = [[get_pcc(hic, hic), pearsonr(hicDecay, hicDecay)[0]]]
        for i in range(len(self.fileList)):
            matrixTmp = np.loadtxt(self.fileList[i])
            if len(matrixTmp) != self.Nbin:
                raise ValueError("Matrix's shape doesn't match!")
            decayTmp = calculate_dis_decay(matrixTmp)
            matrix_simi.append([get_pcc(hic, matrixTmp), pearsonr(hicDecay, decayTmp)[0]])
        return np.array(matrix_simi)    # 1+Nfile

    def evaluate_simi(self):
        compare_list = [read_result(self.hicPath + self.suffix)]
        for i in self.fileList:
            compare_list.append(read_result(i + self.suffix))
        result = Compare(compare_list, self.Nbin, self.suffix, output_path=self.outPath).Output()
        # result = np.r_[np.ones((1, len(result[0]))), result]
        return result   # 1+Nfile

    def evaluate_bound_enrich(self, resultList):
        if not self.bed:
            return np.zeros((len(resultList), 2))
        else:
            return FoldChange(self.bed, resultList, self.resolution, 0, int(self.Nbin * self.resolution), self.chrom, output=self.outPath).Output()

    def evaluate_tad_enrich(self, resultList):
        if not self.bedgraph:
            return np.zeros(len(resultList))
        else:
            return Enrichment(self.bedgraph, resultList, self.resolution, 0, int(self.Nbin * self.resolution), self.chrom, output=self.outPath).Output()


# -*- coding: utf-8 -*-
import argparse, os, re
import numpy as np
import matplotlib

matplotlib.use('Agg')
from matplotlib import pyplot as plt
from scipy.stats import pearsonr, spearmanr
import seaborn as sns


def get_pcc(D_0, D_1, effective_lists=None, verbose=False):
	if D_0.shape != D_1.shape:
		raise ValueError("shape doesn't match")
	n = D_0.shape[0]
	D_0_tmp = D_0.tolist()
	D_1_tmp = D_1.tolist()
	# flatten
	d_0 = []
	d_1 = []
	size = 0
	if effective_lists:
		for i in range(n):
			for j in effective_lists[i]:
				d_0.append(D_0_tmp[i][j])
				d_1.append(D_1_tmp[i][j])
				size += 1
	else:
		d_0 = D_0.flatten().tolist()
		d_1 = D_1.flatten().tolist()
		size = D_0.size
	pcc_ref, pvalue = pearsonr(d_0, d_1)
	# print("pcc_ref=%f, pvalue_ref=%f" % (pcc_ref, pvalue))
	return pcc_ref

def calculate_dis_decay(matrix):
	length = len(matrix)
	decay = [0] * (length - 1)
	for i in range(length - 1):
		count = 0
		sum = 0
		for start in range(length - i - 1):
			count += 1
			if matrix[start][start + i + 1] > 0:
				sum += matrix[start][start + i + 1]
			decay[i] = sum * 1.0 / count
	return decay


def calculateBarcodeDecay(matrix, barcode):
	length = len(matrix)
	decay = [0] * (length - 1)
	count = [0] * (length - 1)
	for i in range(len(barcode)):
		for j in range(i + 1, len(barcode)):
			dis = barcode[j] - barcode[i] - 1
			count[dis] += 1
			decay[dis] += matrix[barcode[i] - 1][barcode[j] - 1]
	for i in range(length - 1):
		if count[i] != 0:
			decay[i] = decay[i] / count[i]
	return decay


def acquire_extn(string):
	filename = os.path.basename(string)
	return os.path.splitext(filename)[0]


def acquireBarcode(matrix, file):
	barcodeDecay = []
	with open(file, "r") as barcode:
		lines = barcode.readlines()
		for num in range(len(lines)):
			if (re.match('>', lines[num]) != None):  # ignore remarks
				info = lines[num].split()[1:]
				info = list(map(int, info))
				print("Barcode: {} with {} coverage".format(info, info[-1] - info[0] + 1))
				barcodeDecay.append(calculateBarcodeDecay(matrix, info))
	return barcodeDecay


def main():
	parser = argparse.ArgumentParser()
	parser.add_argument("-i", dest="matrix", nargs="+", required=True)
	parser.add_argument("-o", dest="output", default="./distance_decay")
	parser.add_argument("-b", dest="barcode", help="specific for LingZhao's result")
	parser.add_argument("-l", dest="length", default=80, type=int)
	parser.add_argument("--list", dest="inputList", action="store_true",
						help="if given, the input is a list of matrix files.")
	parser.add_argument("--name", dest="name", default="iMARGI")
	args = parser.parse_args()

	inputList = []
	matrixList = []
	nameList = []
	decay_list = []
	_N = np.inf  # min count
	if args.inputList:
		for line in open(args.matrix[0], "r"):
			line = line.strip("\n")
			inputList.append(line)
	else:
		inputList = args.matrix
	_M = len(inputList)
	print("Below is the pearsonr btw matrix:")
	for i in range(_M):
		matrixTmp = np.loadtxt(inputList[i])
		matrixList.append(matrixTmp)
		nameList.append(acquire_extn(inputList[i]))
		if _N > len(matrixTmp):
			_N = len(matrixTmp)
		if i != 0:
			print(round(get_pcc(matrixList[0][:_N, :_N], matrixList[i][:_N, :_N]), 4))
	print("Below is the pearsonr btw matrix distance decay:")
	for i in range(_M):
		decay_list.append(calculate_dis_decay(matrixList[i]))
		if i != 0:
			pear, pearPvalue = pearsonr(decay_list[0][:_N - 1], decay_list[i][:_N - 1])
			print(round(pear, 4), round(pearPvalue, 4))

	for i in range(_M):
		print("matrix{}: {}".format(i + 1, nameList[i]))
	plt.figure()
	# color = sns.color_palette("Paired_r", 14)
	color = sns.color_palette("muted", 8)
	print(color)
	try:
		barcodeDecay = acquireBarcode(matrixList[0], args.barcode)
		for i in range(len(barcodeDecay)):
			legend = False
			for j in range(args.length):
				if barcodeDecay[i][j] > 0 and legend == False:
					plt.plot(j + 1, barcodeDecay[i][j], c=color[i + 1], marker='x', label="barcode{}".format(i + 1),
							 markersize=3)
					legend = True
				elif barcodeDecay[i][j] > 0:
					plt.plot(j + 1, barcodeDecay[i][j], c=color[i + 1], marker='x', markersize=3)
		print(list(range(1, args.length + 1)))
		print(decay_list[0][:args.length])
		plt.plot(range(1, args.length + 1), decay_list[0][:args.length], color=color[0],
				 label=nameList[0], linewidth=1.5)
	except:
		print("No input of barcode!")
		for i in range(len(nameList)):
			nameList[0] = "Hi-C contact map"
			nameList[1] = "RAI derived contact map"
			plt.plot(range(1, args.length + 1), decay_list[i][:args.length] / decay_list[i][0], color=color[i],
					 label=nameList[i], linewidth=1.5)
	plt.text(40, 0.2, '**Pearson Correlation={}'.format(round(pear, 4)), fontsize=13)
	plt.xlabel("Distance between two bins (100kb)", size=18)
	plt.ylabel("Average interaction frequency", size=18)
	plt.title("Distance decay (iMARGI-chr10)", size=20)
	plt.legend()
	# plt.xticks([0, 10, 20, 30, 40, 50])
	plt.savefig(args.output + ".pdf")


if __name__ == "__main__":
	main()



import numpy as np
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
import seaborn as sns

def heatmap(matrix=None, out_path=None, boundary=[], diag=False, matrixPath=None, boundaryPath=None):
    if matrixPath:
        matrix = np.loadtxt(matrixPath)
    if boundaryPath:
        boundary = []
        with open(boundaryPath, "r") as result:
            lines = result.readlines()
            for line in lines:
                info = line.split()
                # boundary.append([int(info[0]), int(info[-1])])
                boundary.append([int(info[1]), int(info[5])])
    matrix = matrix.copy()
    plt.figure()
    sns.set()
    cmap = sns.cubehelix_palette(gamma=0.8, as_cmap=True)
    if diag:
        for i in range(len(matrix)):
            matrix[i, i] = 0
    ax = sns.heatmap(matrix, fmt="d", cmap=cmap)
    ax.axis('off')
    if len(boundary):
        ax = add_boundary(ax, boundary, linewidth=1.0)
    if not out_path:
        plt.show()
    else:
        plt.savefig(out_path + ".pdf", format='pdf')

def add_boundary(fig, boundary, color='cornflowerblue', linewidth=0.5):
    for i in range(0, len(boundary), 1):
        x = np.arange(boundary[i][0] - 1, boundary[i][1] + 1)
        y1 = (boundary[i][0] - 1) * np.ones(len(x))
        y2 = (boundary[i][1]) * np.ones(len(x))
        fig.plot(x, y1, color=color, linewidth=linewidth)
        fig.plot(y2, x, color=color, linewidth=linewidth)
        fig.plot(x, y2, color=color, linewidth=linewidth)
        fig.plot(y1, x, color=color, linewidth=linewidth)
    return fig

def read_boundary(file):
    boundary = []
    with open(file, "r") as result:
        lines = result.readlines()
        length = len(lines)
        for line in range(length):
            info = lines[line].split()
            # boundary.append(int(info[1]))   # start bin
            # boundary.append(int(info[5]))   # end bin
            boundary.append([int(info[1]), int(info[5])])
    return np.array(boundary)

def diff_boundary(boundList1, boundList2, ratio=True):
    """
    Calculate the proportion of common bounds from RAI result.
    :param boundList1: Hi-C inferred bounds
    :param boundList2: RAI inferred bounds
    :return:
    """
    bound1 = np.unique(boundList1)
    bound2 = np.unique(boundList2)
    boundCom = []
    boundUniqRAI = []
    for b2 in bound2:
        match = False
        for b1 in bound1:
            if abs(b1 - b2) <= 0:
                match = True
                break
        if match:
            boundCom.append(b2)
        else:
            boundUniqRAI.append(b2)

    if ratio:
        return np.round(len(boundCom) / len(bound2), 4)
    else:
        return boundCom, boundUniqRAI

def diff_domain(boundList1, boundList2):
    """
    Take Hi-C inferred TADs as ref, calculate the proportion of match, merge, split, and shift TADs.
    Roughly define three types of variations of RAI inferred TADs,
    where merge does not generate novel TAD-like domain boundaries,
    while split and shift do.
    :param boundList1: Hi-C inferred TADs
    :param boundList2: RAI inferred TADs
    :return:
    """
    threshold_align = 2
    threshold_shift = 5

    bound1 = np.unique(boundList1)
    bound2 = np.unique(boundList2)
    boundCom = []
    for b2 in bound2:
        match = False
        for b1 in bound1:
            if abs(b1 - b2) < threshold_align:
                match = True
                break
        if match:
            boundCom.append(b2) # boundCom contains all the aligned boundaries of RAI

    matchCount = 0
    mergeCount = 0
    splitCount = 0
    shiftCount = 0
    for tad2 in boundList2:
        if tad2[0] in boundCom:
            if tad2[1] in boundCom:
                state = 0   # candidate merge/match
            else:
                state = 1   # candidate split
        elif tad2[1] in boundCom:
            state = 1
        else:
            state = 2   # candidate shift

        for tad1 in boundList1:
            len1 = tad1[1] - tad1[0] + 1
            len2 = tad2[1] - tad2[0] + 1
            if abs(tad2[0] - tad1[0]) < threshold_align and abs(tad2[1] - tad1[1]) < threshold_align:
                matchCount += 1
                print("{} is a matched TAD with {}".format(tad2, tad1))
                break
            elif state == 2 and \
                    abs(tad2[0] - tad1[0]) < threshold_shift and \
                    abs(tad2[1] - tad1[1]) < threshold_shift:
                shiftCount += 1
                print("{} is a shifted TAD of {}".format(tad2, tad1))
                break
            elif state < 2:
                if abs(tad2[0] - tad1[0]) < threshold_align or abs(tad2[1] - tad1[1]) < threshold_align:
                    if len2 > len1:
                        mergeCount += 1
                        print("{} is a merged TAD from {}".format(tad2, tad1))
                        break
                    elif len1 > len2:
                        splitCount += 1
                        print("{} is a split TAD from {}".format(tad2, tad1))
                        break
    print("#match = {} #merge = {} #split = {} #shift = {}".format(matchCount, mergeCount, splitCount, shiftCount))
    return np.array([matchCount, mergeCount, splitCount, shiftCount]) * (1/len(boundList2))

def construct_dict_from_refgene(refgene, chrom, resolution, minBin, maxBin):
    # construct the hash from refgene, bin->genes
    geneList = [[] for i in range(maxBin - minBin + 1)]
    with open(refgene, "r") as annot:
        lines = annot.readlines()
        for line in lines:
            info = line.split()
            if info[2] == chrom:
                startbin = int(float(info[4]) / resolution)
                endbin = int(float(info[5]) / resolution)
                for i in range(startbin, endbin + 1):
                    if i + 1 >= minBin and i + 1 <= maxBin:
                        geneList[i - minBin + 1].append(info[12])  # official gene symbol
    return geneList

def construct_dict_from_rawdata(refgene, chrom, resolution, minBin, maxBin, mode):
    if mode == "radiclseq":
        gene = 6    # column
    elif mode == "gridseq":
        gene = 3    # column
    geneList = [[] for i in range(maxBin - minBin + 1)]
    with open(refgene, "r") as annot:
        lines = annot.readlines()
        for line in lines:
            info = line.split()
            if info[0] == chrom:
                startbin = int(float(info[1])/resolution)
                endbin = int(float(info[2])/resolution)
                for i in range(startbin, endbin + 1):
                    if i + 1 >= minBin and i + 1 <= maxBin:
                        info[gene] = info[gene].split(".")[0]
                        if info[gene] not in geneList[i - minBin + 1]:
                            geneList[i - minBin + 1].append(info[gene])
    return geneList

def gene_annotation(RAIresultFile, HICresultFile, refgene, outPath="./", mode="imargi"):
    """
    For iMARGI and RIC-seq: refgene;
    For RADICL-seq and GRID-seq: rawdata;
    :param RAIresultFile:
    :param HICresultFile:
    :param refgene:
    :param outPath:
    :return:
    """
    with open(RAIresultFile, "r") as result:
        lines = result.readlines()
        info = lines[0].split()
        chrom = info[0] # chrom
        resolution = int(info[3]) - int(info[2])    # resolution
    boundRAI = read_boundary(RAIresultFile)
    boundHIC = read_boundary(HICresultFile)
    minBin = int(np.min(boundRAI))
    maxBin = int(np.max(boundRAI))
    boundCom, boundUniqRAI = diff_boundary(boundHIC, boundRAI, ratio=False) # list of boundary bins

    if mode == "imargi" or mode == "ricseq":
        geneList = construct_dict_from_refgene(refgene, chrom, resolution, minBin, maxBin)
        suffix = "official_gene_symbol"
    else:
        geneList = construct_dict_from_rawdata(refgene, chrom, resolution, minBin, maxBin, mode)
        if mode == "radiclseq":
            suffix = "ensembml_geneid"
        elif mode == "gridseq":
            suffix = "flybase_geneid"
    # output the gene list
    geneCom = []
    for bound in boundCom:
        if bound != minBin and bound != maxBin:
            geneCom.extend(geneList[bound-minBin])
    geneCom = list(set(geneCom))    # remove the duplicates
    with open(outPath+"geneSet_combound_{}.txt".format(suffix), "w") as ocom:
        ocom.write("\n".join(geneCom))

    geneUniqRAI = []
    for bound in boundUniqRAI:
        if bound != minBin and bound != maxBin:
            geneUniqRAI.extend(geneList[bound-minBin])
    geneUniqRAI = list(set(geneUniqRAI))
    with open(outPath+"geneSet_uniqRAIbound_{}.txt".format(suffix), "w") as ouniq:
        ouniq.write("\n".join(geneUniqRAI))
    print(geneList[0], geneList[-1])

























# _*_ coding: utf-8 _*_
"""
Time:     2022/3/7 15:37
Author:   ZHANG Yuwei
Version:  V 0.1
File:     setup.py
Describe:
"""

import setuptools

# Reads the content of your README.md into a variable to be used in the setup below
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='supertld',
    packages=['supertld'],
    version='0.0.1',
    description='SuperTLD: Detecting TAD-like domains from RNA-associated interactions',
    long_description=long_description,  # loads your README.md
    long_description_content_type="text/markdown",  # README.md is of type 'markdown'
    author='Yu Wei Zhang',
    author_email='ywzhang224@gmail.com',
    url='https://github.com/deepomicslab/SuperTLD',
    classifiers=[  # https://pypi.org/classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],

    download_url="https://github.com/ywZhangV/SuperTLD/archive/refs/tags/v0.0.1.tar.gz",
)


#!/usr/bin/python
import argparse, multiprocessing, os
import numpy as np
import seaborn as sns
import matplotlib

matplotlib.use('Agg')
import matplotlib.pyplot as plt
from scipy import interpolate
from scipy.stats import norm


class globalPar:
	step = 5000  # splitting to small blocks with step size, hic: 5000
	peakbinN = 1  # define the size of peak region through #bin, hic: 3
	backgroundbpSize = 50000  # define the size of background region, hic: 100kb
	maxbpSize = 500000  # consider the max size around the boundary, hic: 500kb


def main():
	print("step={}, peakbinN={}, backgSize={}, maxSize={}".format(globalPar.step, globalPar.peakbinN,
																  globalPar.backgroundbpSize, globalPar.maxbpSize))
	parser = argparse.ArgumentParser()
	parser.add_argument("--bed", dest="bed", nargs="+", required=True, help="Epigenome bed file, eg. CTCF")
	parser.add_argument("-i", dest="result", nargs="+", required=True, help="TADcaller result file in 8col format")
	parser.add_argument("-r", dest="resolution", type=int, required=True)
	parser.add_argument("-c", dest="chrom", required=True)
	parser.add_argument("-o", dest="output", default=None)
	parser.add_argument("--plot", dest="plot", action="store_true", help="if given, draw the bound profile plot")
	parser.add_argument("-s", dest="start", type=int, default=0, help="the start bp position")
	parser.add_argument("-e", dest="end", type=int, default=np.inf, required=True, help="the end bp position")
	parser.add_argument("--common", dest="common", action="store_true", help="if given, evaluate the common part of two results")
	parser.add_argument("--list", dest="list", action="store_true", help="if given, the input is a list of results")
	parser.add_argument("--suffix", dest="suffix", default=".multi2D_AllH2.tsv", help="default: .binary.original.tsv")
	args = parser.parse_args()

	if args.output == None:
		output = os.path.split(os.path.realpath(args.result[0]))[0]
	else:
		output = args.output
	if args.list == True:
		resultList = []
		for line in open(args.result[0], "r"):
			line = line.strip("\n")
			resultList.append(line + args.suffix)
		FoldChange(args.bed, resultList, args.resolution, args.start, args.end, args.chrom, output, args.plot,
				   args.common)
	else:
		resultList = []
		for result in args.result:
			resultList.append(result + args.suffix)
		FoldChange(args.bed, resultList, args.resolution, args.start, args.end, args.chrom, output, args.plot,
				   args.common)


class FoldChange():
	def __init__(self, bedFILE, resultFILE, resolution, start, end, chr, output=None, plot=False, common=False):
		self.chr = chr
		self.name = resultFILE[0]
		self.start = start - globalPar.maxbpSize
		self.end = end + globalPar.maxbpSize
		# self.step = int(0.1 * resolution)
		self.step = globalPar.step
		self.profile = []
		for i in range(len(bedFILE)):
			self.profile.append(self.read_profile(bedFILE[i]))
		self.resolution = resolution
		self.peak = np.zeros(int((globalPar.peakbinN * self.resolution) / self.step))
		# self.peak = np.zeros(int(2 * gap / self.step))
		self.background = np.zeros(int(globalPar.backgroundbpSize / self.step))
		self.output = os.path.dirname(output)
		self.plot = plot
		self.result = []
		for i in range(len(resultFILE)):
			self.result.append(self.read_result(resultFILE[i]))
			print("{}th result with {} boundaries".format(i, len(self.result[i])))
		self.common = common
		if self.common:	# only two result files
			self.boundB = []
			self.result = self.deriveCommon(self.result)
		self.plotprofile = None
		self.foldchangeResult = np.zeros((len(self.result), len(self.profile) * 2))
		self.profileTmp = None
		if self.plot:
			self.run()
		else:
			self.run_fast()

	def run_fast(self):
		for i in range(len(self.profile)):
			self.profileTmp = self.profile[i]
			pool = multiprocessing.Pool()
			result = pool.map(self.calculate_fold, self.result)
			pool.close()
			pool.join()
			for j in range(len(self.result)):
				self.foldchangeResult[j, i*2], self.foldchangeResult[j, 2*i+1] = result[j]
		# np.savetxt(self.output + "/enrichment_" + self.chr + "foldchange.txt", self.foldchangeResult, fmt='%.4f')

	def run(self):
		# color = sns.color_palette("Paired_r", 14)
		color = sns.color_palette("muted", 8)
		xaxis = list(range(np.negative(globalPar.maxbpSize), globalPar.maxbpSize + self.resolution, self.step))
		xnew = np.linspace(np.min(xaxis), np.max(xaxis), 1000)
		for i in range(len(self.profile)):
			self.profileTmp = self.profile[i]
			if self.plot:
				sns.set(style="ticks")
				fig, ax = plt.subplots(figsize=[5, 3.5])
			# label = ["Hi-C only: 0.065", "iMARGI only: 0.875", "Common: 2.545"]
			label = ["TAD uniq: 0.014", "TLD uniq: 0.1433", "Common: 0.3115"]
			for j in range(len(self.result)):
				self.foldchangeResult[j, 2 * i], self.foldchangeResult[j, 2 * i + 1] = self.calculate_fold(self.result[j])
				if self.plot:
					func = interpolate.interp1d(xaxis, self.plotprofile / len(self.result[j]), kind='cubic')
					if self.common:
						sns.lineplot(xnew, func(xnew), color=color[j], label=label[j])
					else:
						sns.lineplot(xnew, func(xnew), color=color[j], label="label_{}".format(j))
			# fig.subplots_adjust(right=0.75, top=0.8, bottom=0.2)
			if self.plot:
				plt.ylabel("Number of peaks per {}kb".format(int(self.step / 1000)), fontsize=16)
				plt.xlabel("")
				plt.title("GRID-seq-chr2R", size=20)
				bonus = int(0.5*self.resolution)
				plt.xticks([-400000, -200000, bonus, 200000+self.resolution, 400000+self.resolution],
						   [r"-400kb", r"-200kb", r"boundary", r"200kb", r"400kb"])
				plt.tight_layout()
				plt.savefig(self.output + "/enrichment_profile" + str(i) + "_" + self.chr + ".pdf")
				if self.common:
					fig, ax = plt.subplots(figsize=[5, 3.5])
					color1 = sns.color_palette("Paired_r", len(self.boundB))
					result = []
					for j in range(len(self.boundB)):
						result.append(self.calculate_fold([self.boundB[j]]))
						if self.plot:
							func = interpolate.interp1d(xaxis, self.plotprofile, kind='cubic')
							sns.lineplot(xnew, func(xnew), color=color1[j], label='result' + str(j))

					plt.ylabel("Number of peaks per {}kb".format(int(self.step / 1000)), fontsize=18)
					plt.xlabel("")
					plt.tight_layout()

					plt.savefig(self.output + "/enrichment_profile_boundB" + "_" + self.chr + ".pdf")
					np.savetxt(self.output + "/enrichment_" + self.chr + "boundB_foldchange.txt",
							   np.array(np.concatenate([self.boundB, result], axis=1), dtype=np.str), fmt="%s")
		# print(self.foldchangeResult)
		np.savetxt(self.output + "/enrichment_" + self.chr + "foldchange.txt", self.foldchangeResult, fmt='%.4f')

	def Output(self):
		return self.foldchangeResult

	def read_profile(self, bedFILE):
		binN = (self.end - self.start) / self.step
		profile = np.zeros(int(binN))
		with open(bedFILE, "r") as bed:
			lines = bed.readlines()
			for peak in lines:
				peak = peak.split()
				if peak[0] == self.chr and int(peak[1]) >= self.start and int(peak[2]) <= self.end:
					left = int((int(peak[1]) - self.start) / self.step)
					right = int((int(peak[2]) - self.start) / self.step)
					if left == right:
						profile[left] += 1
					elif right - left == 1:
						length = int(peak[2]) - int(peak[1]) + 1
						left_part = self.start + self.step * right - int(peak[1])
						profile[left] += left_part / length
						profile[right] += 1 - (left_part / length)
					# print("right=left+1, ", peak, left, right, length, left_part)
					else:
						print(peak, "stoped.")
						exit()
				elif peak[0] == self.chr:
					if int(peak[2]) > self.start and int(peak[2]) <= self.end:
						length = int(peak[2]) - int(peak[1]) + 1
						pos = int((int(peak[2]) - self.start) / self.step)
						if pos == 0:
							profile[pos] += (int(peak[2]) - self.start) / length
						elif pos == 1:
							profile[0] += self.step / length
							profile[1] += (int(peak[2]) - self.start - self.step) / length
						else:
							print(peak, "stoped.")
							exit()
					elif int(peak[1]) >= self.start and int(peak[1]) < self.end:
						length = int(peak[2]) - int(peak[1]) + 1
						pos = int((int(peak[2]) - self.start) / self.step)
						if binN == pos + 1:
							profile[pos] += (self.end - int(peak[1])) / length
						elif binN == pos + 2:
							profile[pos] += (self.end - self.step - int(peak[1])) / length
							profile[pos + 1] += self.step / length
						else:
							print(peak, "stoped.")
							exit()
		return profile

	def read_result(self, resultFILE):
		"""
		Derive all the boundaries except the start and end position (allow duplication).
		:param resultFILE:
		:return: list of bounds, bound: [start pos, end pos]
		"""
		start = self.start + globalPar.maxbpSize
		end = self.end - globalPar.maxbpSize
		bound_list = []
		result = np.loadtxt(resultFILE, np.str)
		for i in range(len(result)):
			bound1 = int(result[i][2])
			bound2 = int(result[i][7])
			if bound1 != start and bound1 != end:
				bound_list.append([bound1, bound1 + self.resolution])
			if bound2 != start and bound2 != end:
				bound_list.append([bound2 - self.resolution, bound2])

		return bound_list

	def calculate_fold(self, result):
		self.peak = np.zeros(int(globalPar.peakbinN * self.resolution / self.step))
		self.background = np.zeros(int(globalPar.backgroundbpSize / self.step))
		self.plotprofile = np.zeros(int((2 * globalPar.maxbpSize+self.resolution) / self.step))
		for i in result:
			self.process_tad(i, self.profileTmp)
		peak_value = np.mean(self.peak / len(result))
		background_profile = self.background / (len(result) * 2)
		background_value = np.mean(background_profile)
		foldChange = (peak_value / background_value) - 1  # important!
		# calculate pvalue
		pvalue_of_peak = 1 - norm.cdf(peak_value, loc=background_value, scale=np.std(background_profile))
		if len(result) == 1:
			print(result, foldChange, pvalue_of_peak)
		return foldChange, pvalue_of_peak

	def process_tad(self, bound, profile):
		start_cor = int(((bound[0] - (globalPar.peakbinN - 1)*0.5*self.resolution) - self.start) / self.step)
		back_cor = int(((bound[0] - globalPar.maxbpSize) - self.start) / self.step)
		back_cor_2 = int(((bound[1] + globalPar.maxbpSize - globalPar.backgroundbpSize) - self.start) / self.step)
		for i in range(len(self.plotprofile)):
			self.plotprofile[i] += profile[back_cor + i]
			if i < len(self.background):
				# print(bound, start_cor, back_cor, back_cor_2, i)
				self.background[i] += profile[back_cor + i]
				self.background[i] += profile[back_cor_2 + i]
			if i < len(self.peak):
				self.peak[i] += profile[start_cor + i]

	def deriveCommon(self, resultList):
		"""
		Derive three lists of bounds from two derived lists of TAD bounds (not allow duplication).
		:param resultList: two derived lists of TAD bounds
		:return:
		"""
		boundA = []
		boundB = []
		boundCommon = []
		for b1 in resultList[0]:
			match = False
			for b2 in resultList[1]:
				if abs(b1[0] - b2[0]) / self.resolution <= 0:
					match = True
					# print("b1={}, b2={}".format(b1, b2))
					break
			if match:
				boundCommon.append(b1)
			else:
				boundA.append(b1)
		for b2 in resultList[1]:
			match = False
			for b1 in resultList[0]:
				if abs(b1[0] - b2[0]) / self.resolution <= 0:
					match = True
					# print("b1={}, b2={}".format(b1, b2))
					break
			if not match:
				boundB.append(b2)
		# boundA = [i for i in resultList[0] if i not in resultList[1]]
		# boundB = [i for i in resultList[1] if i not in resultList[0]]
		# boundCommon = [i for i in resultList[0] if i in resultList[1]]
		boundA = self.removedup(boundA)
		boundB = self.removedup(boundB)
		boundCommon = self.removedup(boundCommon)
		print("Specific Bound for Hi-C: ", len(boundA))
		print("Specific Bound for B: ", len(boundB))
		print("Common Bound: ", len(boundCommon))
		for i in boundB:
			if not i in self.boundB:
				self.boundB.append(i)
		return [boundA, boundB, boundCommon]

	def removedup(self, list):
		new_list = []
		for i in list:
			if not i in new_list:
				new_list.append(i)
		return new_list

if __name__ == "__main__":
	main()



# _*_ coding: utf-8 _*_
"""
Time:     2021/12/14 21:50
Author:   ZHANG Yuwei
Version:  V 0.1
File:     KR_norm_juicer.py
Describe:
"""

import numpy as np

def KRnorm_asym(test, verbose=False, oneNorm=False):
    """
    Construct symMatrix = [[0, A], [A.T, 0]]
    :param test: input asymmetric matrix
    :param verbose:
    :param oneNorm:
    :return: norm vector: result, shape = (M+N,);
    norm matrix: test, same shape as input
    """
    # construct sym matrix, shape = (M+N, M+N)
    length_r = test.shape[0]
    length_c = test.shape[1]
    new_test = np.zeros((length_r + length_c, length_r + length_c))
    new_test[:length_r, length_r:] = test
    new_test[length_r:, :length_r] = test.T
    # add some noise
    noise_matrix = np.random.normal(loc=1e-5, scale=1e-6, size=(length_r + length_c, length_r + length_c))
    noise_matrix = np.triu(noise_matrix) + np.triu(noise_matrix, 1).T
    # print(noise_matrix, np.min(noise_matrix), np.max(noise_matrix))
    new_test += noise_matrix
    new_test[new_test < 0] = 0
    result = Compute_KR(new_test, asy=True, Onenorm=oneNorm, verbose=verbose).run()
    result_r = result[:length_r]
    result_c = result[length_r:]
    for i in range(length_r):
        for j in range(length_c):
            if not (np.isnan(result_r[i])) and not (np.isnan(result_c[j])):
                test[i, j] = test[i, j] / (result_r[i] * result_c[j])
    return test, result

def KRnorm_sym(test, verbose=False, oneNorm=False):
    result = Compute_KR(test, verbose=verbose, Onenorm=oneNorm).run()
    length = len(test)
    for i in range(length):
        for j in range(i, length, 1):
            if (not (np.isnan(result[i])) and not (np.isnan(result[j]))):
                test[i, j] = test[i, j] / (result[i] * result[j])
                test[j, i] = test[i, j]
    return test, result

class Compute_KR():
    def __init__(self, matrix, asy=False, verbose=False, Onenorm=False):
        self.matrix = matrix
        self.length = len(self.matrix)
        self.asy = asy
        self.valid_thre = None
        self.verbose = verbose
        self.norm = Onenorm

    def run(self):
        if self.verbose:
            print("matrix's shape = {}".format(self.matrix.shape))
        norm = self.computeKR()
        factor = self.getSumFactor(norm)
        if self.verbose:
            print("norm", norm)
            print("factor", factor)
        if self.norm:
            factor = 1
        return np.asarray(norm) * factor

    def computeKR(self):
        recalculate = True
        offset = self.getOffset(0.0)
        iteration = 1
        max_iter = 60
        while (recalculate and iteration <= max_iter):
            sparsematrix = self.populateMatrix(self.matrix, offset)  # delete the rows/cols according to offset
            new_size = len(sparsematrix)
            if self.verbose:
                print("iteration = {}, total_bin = {}, left_bin = {}".format(iteration, self.length, new_size))
            x0 = [1.0] * new_size
            x0 = self.computeKRNormVector(sparsematrix, 1e-6, x0, 0.1)
            recalculate = False
            rowsTossed = 0
            if (x0 is None or iteration == max_iter-1):
                recalculate = True
                if (iteration < max_iter-1):    #x0 is None: not converse within 100 times
                    offset = self.getOffset(iteration)  # increase the threshold to delete more data
                else:   #iteration==max_iter-1
                    offset = self.getOffset(max_iter+5)
                    if self.verbose:
                        print(iteration, "Changed threshold by 5%.")
            else:   #x0 is valid
                kr = np.zeros(self.length)
                for i in range(self.length):
                    if offset[i] == -1:
                        kr[i] = np.nan
                    else:
                        kr[i] = 1.0 / x0[offset[i]]  # transform the factor
                if self.verbose:
                    print(iteration, kr)
                    print("nan has {}, zero has {}".format(len(np.where(np.isnan(kr)==True)[0]), len(np.where(kr==0)[0])))
                mySum = self.getSumFactor(kr)  # an overall factor to the whole matrix
                index = 0
                if self.valid_thre == None:
                    if self.asy:
                        test = kr * mySum
                        self.valid_thre = np.min(test[np.where(np.isnan(test)==False)])
                    else:
                        self.valid_thre = 0.01
                if self.verbose:
                    print(self.valid_thre, "valid_threshold")
                for i in range(self.length):
                    if kr[i] * mySum < self.valid_thre:  # factor is valid but too little
                        offset[i] = -1  # delete the row and calculate again
                        rowsTossed += 1
                        recalculate = True
                    elif (offset[i] != -1):  # valid
                        offset[i] = index
                        index += 1
            if self.verbose:
                print("Finish Iteration = {}, recalculate: {}".format(iteration, recalculate))
            iteration += 1
        if (iteration > max_iter and recalculate):
            if self.verbose:
                print("Iteration > {}, output all nan.".format(max_iter))
            kr = [np.nan] * self.length
        return kr

    def getOffset(self, percent=0.0):
        """
        set rows/cols that cannot reach the threshold to -1 in offset list
        :param matrix: input matrix
        :return: offset list (all deleted rows are set to -1)
        """
        rowSums = np.zeros(self.length)
        for i in range(self.length):
            rowSums[i] = np.sum(self.matrix[i,])
        thresh = 0.0
        if (percent != 0.0):
            no_zeros = np.where(rowSums != 0)
            posRowSums = np.sort(rowSums[no_zeros])
            thresh = posRowSums[int(len(posRowSums) * percent / 100)]
        if self.verbose:
            print(thresh)
        offset = [0] * self.length
        index = 0
        for i in range(self.length):
            if (np.sum(self.matrix[i,]) <= thresh):
                offset[i] = -1
            else:
                offset[i] = index
                index += 1
        return offset

    def populateMatrix(self, matrix, offset):
        dele_row = np.where(np.asarray(offset) == -1)[0]
        matrix = np.delete(matrix, dele_row, axis=0)
        matrix = np.delete(matrix, dele_row, axis=1)
        return matrix

    def computeKRNormVector(self, matrix, tol, x0, delta=0.1, Delta=3):
        """
            A balancing algorithm for symmetric matrics, which attempts to find a vector x0 such that
            diag(x0)*A*diag(x0) is close to doubly stochastic.

            :param matrix: input matrix, must be symmetric and nonnegative.
            :param tol: error tolerance
            :param x0: initial guess, default: ones
            :param delta/Delta: how close/far balancing vectors can get to/from the edge of the positive cone.
            :return: balancing vector x0.
	    """
        n = len(x0)  # No. selected rows
        e = [1.0] * n   #initialization
        # Inner stopping criterion parameters
        g = 0.9
        etamax = 0.1
        eta = etamax
        rt = pow(tol, 2)  # default: 1e-12
        v = x0 * (matrix.dot(x0))  # shape = (509,)
        rk = 1.0 - v
        rho_km1 = rk.transpose().dot(rk)  # a scaler
        rout = rold = rho_km1
        MVP = 0  # We will count matrix vector products
        not_changing = 0  # Outer iteration count
        # Nepoch = 0
        # while (rout > rt and not_changing < 100 and Nepoch < 50000):
        while rout > rt and not_changing < 100:
            k = 0
            y = np.copy(e)
            rho_km2 = rho_km1
            innertol = max(pow(eta, 2) * rout, rt)
            while (rho_km1 > innertol):  # inner iteration by CG
                k += 1
                if (k == 1):
                    Z = rk / v
                    p = np.copy(Z)
                    rho_km1 = rk.transpose().dot(Z)
                else:
                    beta = rho_km1 / rho_km2
                    p = Z + beta * p
                if k > 10:
                    break
                # update search direction efficiently.
                tmp = matrix.dot(x0 * p)
                w = x0 * tmp + v * p
                alpha = rho_km1 / (p.transpose().dot(w))
                ynew = y + alpha * p

                # test distance to boundary of cone
                minynew = np.amin(ynew)
                if (minynew <= delta):
                    if (delta == 0):
                        break
                    gamma = np.inf
                    for i in range(len(ynew)):
                        if (alpha * p[i] < 0 and (delta - y[i]) / (alpha * p[i]) < gamma):
                            gamma = (delta - y[i]) / (alpha * p[i])
                    y += gamma * alpha * p
                    break

                maxynew = np.amax(ynew)
                if (maxynew >= Delta):
                    gamma = np.inf
                    for i in range(len(ynew)):
                        if (ynew[i] > Delta and (Delta - y[i]) / (alpha * p[i]) < gamma):
                            gamma = (Delta - y[i]) / (alpha * p[i])
                    y += gamma * alpha * p
                    break

                rho_km2 = rho_km1
                y = np.copy(ynew)
                rk -= alpha * w
                Z = rk / v
                rho_km1 = rk.transpose().dot(Z)
                if self.verbose:
                    print("-----------------")
            x0 = x0 * y
            v = x0 * matrix.dot(x0)
            rk = 1.0 - v    # the error of balancing, obj: zeros
            rho_km1 = rk.transpose().dot(rk)
            if self.verbose:
                print(rho_km1, rout, abs(rho_km1 - rout))
                print(n, "x0 has nan {}, y has nan {}".format(len(np.where(np.isnan(np.array(x0))==True)[0]), len(np.where(np.isnan(np.array(y))==True)[0])))
            if abs(rho_km1 - rout) < 1e-6 or np.isinf(rho_km1):
                not_changing += 1
                # Nepoch = 0
            # else:
            #     Nepoch += 1
            rout = rho_km1
            MVP += k + 1
            # Update inner iteration stopping criterion
            rat = rout / rold
            rold = rout
            r_norm = np.sqrt(rout)
            eta_o = eta
            eta = g * rat
            if (g * pow(eta_o, 2) > 0.1):
                eta = max(eta, g * pow(eta_o, 2))
            eta = max(min(eta, etamax), 0.5 * tol / r_norm)
            if self.verbose:
                # print("not_changing time = {}, current Nepoch = {}".format(not_changing, Nepoch))
                print("not_changing time = {}".format(not_changing))
        if not_changing >= 100 or len(np.where(np.isnan(np.array(x0))==True)[0]) == n:
            return None
        return x0

    def getSumFactor(self, norm_list):
        """
		Use the calculated bin factors to calculate the overall factor to the whole normalized matrix.
	    :param norm_list: the calculated factor list
	    :return: an overall factor to the whole matrix
	    """
        matrix_sum = 0  # raw pixel sum
        norm_sum = 0  # normalized pixel sum
        for i in range(self.length):
            for j in range(i, self.length, 1):
                if (not (np.isnan(norm_list[i])) and not (np.isnan(norm_list[j])) and norm_list[i] > 1e-36 and norm_list[j] > 1e-36):  # only if the factor is larger than 0, valid
                    if (i == j):  # pixel on the diagonal
                        norm_sum += self.matrix[i, j] / (norm_list[i] * norm_list[j])
                        matrix_sum += self.matrix[i, j]
                    else:
                        norm_sum += (2 * self.matrix[i, j]) / (norm_list[i] * norm_list[j])
                        matrix_sum += 2 * self.matrix[i, j]
        if self.verbose:
            print(norm_sum, matrix_sum)
        return np.sqrt(norm_sum / matrix_sum)


